from typing import Any, Dict, List, Optional, Set, Union

from pydantic import Extra, Field, root_validator, validator

from aiosox.utils import BaseModel

from .enums import (
    JSONReference,
)

SPECIAL_PATH_FORMAT: str = "#-special-path-#-{}-#-special-#"


class JsonSchemaObject(BaseModel):
    __constraint_fields__: Set[str] = {
        "exclusiveMinimum",
        "minimum",
        "exclusiveMaximum",
        "maximum",
        "multipleOf",
        "minItems",
        "maxItems",
        "minLength",
        "maxLength",
        "pattern",
    }
    # __extra_key__: str = SPECIAL_PATH_FORMAT.format("extras")

    @root_validator(pre=True)
    def validate_exclusive_maximum_and_exclusive_minimum(cls, values: Dict[str, Any]) -> Any:
        exclusive_maximum: Union[float, bool, None] = values.get("exclusiveMaximum")
        exclusive_minimum: Union[float, bool, None] = values.get("exclusiveMinimum")

        if exclusive_maximum is True:
            values["exclusiveMaximum"] = values["maximum"]
            del values["maximum"]
        elif exclusive_maximum is False:
            del values["exclusiveMaximum"]
        if exclusive_minimum is True:
            values["exclusiveMinimum"] = values["minimum"]
            del values["minimum"]
        elif exclusive_minimum is False:
            del values["exclusiveMinimum"]
        return values

    @validator("ref")
    def validate_ref(cls, value: Any) -> Any:
        if isinstance(value, str) and "#" in value:
            if value.endswith("#/"):
                return value[:-1]
            elif "#/" in value or value[0] == "#" or value[-1] == "#":
                return value
            return value.replace("#", "#/")
        return value

    items: Union[List["JsonSchemaObject"], "JsonSchemaObject", None] = None
    # uniqueItem: Optional[bool]
    type: Union[str, List[str], None] = None
    format: Optional[str] = None
    # pattern: Optional[str]
    minLength: Optional[int] = None
    maxLength: Optional[int] = None
    minimum: Optional[float] = None
    maximum: Optional[float] = None
    minItems: Optional[int] = None
    maxItems: Optional[int] = None
    # multipleOf: Optional[float]
    # exclusiveMaximum: Union[float, bool, None]
    # exclusiveMinimum: Union[float, bool, None]
    additionalProperties: Union["JsonSchemaObject", bool, None] = None
    # patternProperties: Optional[Dict[str, "JsonSchemaObject"]]
    oneOf: Optional[List["JsonSchemaObject"]] = None
    anyOf: Optional[List["JsonSchemaObject"]] = None
    allOf: Optional[List["JsonSchemaObject"]] = None
    enum: Optional[List[Any]] = None
    # writeOnly: Optional[bool]
    properties: Optional[Dict[str, "JsonSchemaObject"]] = None
    required: List[str] = []
    ref: Optional[str] = Field(default=None, alias="$ref")
    nullable: Optional[bool] = False
    # x_enum_varnames: List[str] = Field(default=[], alias="x-enum-varnames")
    description: Optional[str] = None
    title: Optional[str] = None
    example: Any
    examples: Any
    default: Optional[Any] = None
    id: Optional[str] = Field(default=None, alias="$id")
    custom_type_path: Optional[str] = Field(default=None, alias="customTypePath")

    extras: Dict[str, Any] = Field(default=None)

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data: Any) -> None:  # type: ignore
        super().__init__(**data)
        self.extras = {k: v for k, v in data.items() if k not in EXCLUDE_FIELD_KEYS}

    def is_object(self) -> bool:
        return (
            self.properties is not None
            or self.type == "object"
            and not self.allOf
            and not self.oneOf
            and not self.anyOf
            and not self.ref
        )

    def is_array(self) -> bool:
        return self.items is not None or self.type == "array"

    def ref_object_name(self) -> str:  # pragma: no cover
        return self.ref.rsplit("/", 1)[-1]  # type: ignore

    @validator("items", pre=True)
    def validate_items(cls, values: Any) -> Any:
        # this condition expects empty dict
        return values or None

    def has_default(self) -> bool:
        return "default" in self.__fields_set__

    def has_constraint(self) -> bool:
        return bool(self.__constraint_fields__ & self.__fields_set__)

    def ref_type(self) -> Optional[JSONReference]:
        if self.ref:
            return JSONReference(self.ref)
        return None  # pragma: no cover


class ApiSchema(JsonSchemaObject):

    discriminator: Optional[str] = None

    class Config:
        extra = Extra.forbid


DEFAULT_FIELD_KEYS: Set[str] = {
    "example",
    "examples",
    "description",
    "title",
}

EXCLUDE_FIELD_KEYS = (set(JsonSchemaObject.__fields__) - DEFAULT_FIELD_KEYS) | {
    "$id",
    "$ref",
    # JsonSchemaObject.__extra_key__,
}



from typing import List, Optional, Union

from pydantic import Extra

from aiosox.utils import BaseModel

from .bindings import (
    HttpOperationBinding,
    KafkaOperationBinding,
    WebSocketsOperationBinding,
)
from .message import ApiMessage
from .reference import Reference
from .tag import ApiTag


class OperationBindings(BaseModel):
    """
    Map describing protocol-specific definitions for a operation.
    """

    http: Optional[HttpOperationBinding] = None
    """
    Protocol-specific information for an HTTP operation.
    """

    socketio: Optional[WebSocketsOperationBinding] = None
    """
    Protocol-specific information for a WebSockets operation.
    """

    kafka: Optional[KafkaOperationBinding] = None
    """
    Protocol-specific information for a Kafka operation.
    """


class OperationTrait(BaseModel):
    """
    Describes a trait that MAY be applied to an Operation Object. This object MAY contain any
    property from the Operation Object, except message and traits.
    """

    operationId: Optional[str] = None
    """
    Unique string used to identify the operation. The id MUST be unique among all operations described
    in the API. The operationId value is case-sensitive. Tools and libraries MAY use the operationId to
    uniquely identify an operation, therefore, it is RECOMMENDED to follow common programming naming
    conventions.
    """

    summary: Optional[str] = None
    """
    A short summary of what the operation is about.
    """

    description: Optional[str] = None
    """
    A verbose explanation of the operation. CommonMark syntax can be used for rich text representation.
    """

    tags: Optional[List[ApiTag]] = None
    """
    A list of tags for API documentation control. Tags can be used for logical grouping of operations.
    """

    bindings: Optional[Union[OperationBindings, Reference]] = None
    """
    A map where the keys describe the name of the protocol and the values describe protocol-specific
    definitions for the operation.
    """

    class Config:
        extra = Extra.forbid
        schema_extra = {"examples": [{"bindings": {"amqp": {"ack": False}}}]}


class ApiOperation(BaseModel):
    """Describes a publish or a subscribe operation. This provides a place to document
    how and why messages are sent and received.
    For example, an operation might describe a chat application use case where a user
    sends a text message to a group. A publish operation describes messages that are
    received by the chat application, whereas a subscribe operation describes messages
    that are sent by the chat application."""

    operationId: Optional[str] = None
    """
    Unique string used to identify the operation. The id MUST be unique among all
    operations described in the API. The operationId value is case-sensitive. Tools and
    libraries MAY use the operationId to uniquely identify an operation, therefore, it
    is RECOMMENDED to follow common programming naming conventions.
    """

    summary: Optional[str] = None
    """
    A short summary of what the operation is about.
    """

    description: Optional[str] = None
    """
    A verbose explanation of the operation. CommonMark syntax can be used for rich
    text representation.
    """

    tags: Optional[List[ApiTag]] = None
    """
    A list of tags for API documentation control. Tags can be used for logical grouping
    of operations.
    """

    bindings: Optional[Union[OperationBindings, Reference]] = None
    """
    A map where the keys describe the name of the protocol and the values describe
    protocol-specific definitions for the operation.
    """

    traits: Optional[List[Union[OperationTrait, Reference]]] = None
    """
    A list of traits to apply to the operation object. Traits MUST be merged into the
    operation object using the JSON Merge Patch algorithm in the same order they are
    defined here.
    """

    message: Optional[Union[ApiMessage, Reference]] = None
    """
    A definition of the message that will be published or received by this operation.
    Map containing a single oneOf key is allowed here to specify multiple messages.
    However, a message MUST be valid only against one of the message objects.
    """

    class Config:
        extra = Extra.forbid
        schema_extra = {
            "examples": [
                {
                    "operationId": "registerUser",
                    "summary": "Action to sign a user up.",
                    "description": "A longer description",
                    "tags": [{"name": "user"}, {"name": "signup"}, {"name": "register"}],
                    "message": {
                        "headers": {
                            "type": "object",
                            "properties": {
                                "applicationInstanceId": {
                                    "description": "Unique identifier for a given instance of the publishing application",
                                    "type": "string",
                                }
                            },
                        },
                        "payload": {
                            "type": "object",
                            "properties": {
                                "user": {"$ref": "#/components/schemas/userCreate"},
                                "signup": {"$ref": "#/components/schemas/signup"},
                            },
                        },
                    },
                    "bindings": {"amqp": {"ack": False}},
                    "traits": [{"$ref": "#/components/operationTraits/kafka"}],
                }
            ]
        }



import pytest

from aiosox.sio import server


class Test_Socketioserver_Register_namespace:
    @pytest.fixture()
    def socketioserver(self):
        return server.SocketIoServer()

    def test_register_namespace_1(self, socketioserver):
        socketioserver.register_namespace()



from typing import Any, Dict, List, Mapping, Optional, Sequence, Set, Type, Union

from fastapi.encoders import jsonable_encoder
from pydantic import AnyUrl, Extra, Field, constr, root_validator, validator

from aiosox.utils import BaseModel

from .bindings import ApiServerBindings, ChannelBindings
from .channel import ApiChannel
from .message import ApiMessage, CorrelationId, MessageBindings, MessageTrait
from .operation import OperationBindings, OperationTrait
from .parameter import ApiParameter
from .reference import Reference
from .schema import ApiSchema
from .security import ApiSecurityScheme
from .server import ApiServer

ComponentKey = str


class ApiComponents(BaseModel):
    """
    Holds a set of reusable objects for different aspects of the AsyncAPI specification. All
    objects defined within the components object will have no effect on the API unless they
    are explicitly referenced from properties outside the components object.
    """

    schemas: Optional[Dict[ComponentKey, Union[ApiSchema, Reference]]] = None
    """
    An object to hold reusable Schema Objects.
    """

    servers: Optional[Dict[ComponentKey, Union[ApiServer, Reference]]] = None
    """
    An object to hold reusable Server Objects.
    """

    channels: Optional[Dict[ComponentKey, ApiChannel]] = None
    """
    An object to hold reusable Channel Item Objects.
    """

    messages: Optional[Dict[ComponentKey, Union[ApiMessage, Reference]]] = None
    """
    An object to hold reusable Message Objects.
    """

    security_schemes: Optional[Dict[ComponentKey, Union[ApiSecurityScheme, Reference]]] = None
    """
    An object to hold reusable Security Scheme Objects.
    """

    parameters: Optional[Dict[ComponentKey, Union[ApiParameter, Reference]]] = None
    """
    An object to hold reusable Parameter Objects.
    """

    correlationIds: Optional[Dict[ComponentKey, Union[CorrelationId, Reference]]] = None
    """
    An object to hold reusable Correlation ID Objects.
    """

    operationTraits: Optional[Dict[ComponentKey, Union[OperationTrait, Reference]]] = None
    """
    An object to hold reusable Operation Trait Objects.
    """

    messageTraits: Optional[Dict[ComponentKey, Union[MessageTrait, Reference]]] = None
    """
    An object to hold reusable Message Trait Objects.
    """

    serverBindings: Optional[Dict[ComponentKey, Union[ApiServerBindings, Reference]]] = None
    """
    An object to hold reusable Server Bindings Objects.
    """

    channelBindings: Optional[Dict[ComponentKey, Union[ChannelBindings, Reference]]] = None
    """
    An object to hold reusable Channel Bindings Objects.
    """

    operationBindings: Optional[Dict[ComponentKey, Union[OperationBindings, Reference]]] = None
    """
    An object to hold reusable Operation Bindings Objects.
    """

    messageBindings: Optional[Dict[str, Union[MessageBindings, Reference]]] = None
    """
    An object to hold reusable Message Bindings Objects.
    """

    class Config:
        extra = Extra.forbid
        schema_extra = {
            "examples": [
                {
                    "components": {
                        "schemas": {
                            "Category": {
                                "type": "object",
                                "properties": {
                                    "id": {"type": "integer", "format": "int64"},
                                    "name": {"type": "string"},
                                },
                            },
                            "Tag": {
                                "type": "object",
                                "properties": {
                                    "id": {"type": "integer", "format": "int64"},
                                    "name": {"type": "string"},
                                },
                            },
                        },
                        "servers": {
                            "development": {
                                "url": "development.gigantic-server.com",
                                "description": "Development server",
                                "protocol": "amqp",
                                "protocolVersion": "0.9.1",
                            }
                        },
                        "channels": {
                            "user/signedup": {
                                "subscribe": {
                                    "message": {"$ref": "#/components/messages/userSignUp"}
                                }
                            }
                        },
                        "messages": {
                            "userSignUp": {
                                "summary": "Action to sign a user up.",
                                "description": "Multiline description of what this action does.\nHere you have another line.\n",
                                "tags": [{"name": "user"}, {"name": "signup"}],
                                "headers": {
                                    "type": "object",
                                    "properties": {
                                        "applicationInstanceId": {
                                            "description": "Unique identifier for a given instance of the publishing application",
                                            "type": "string",
                                        }
                                    },
                                },
                                "payload": {
                                    "type": "object",
                                    "properties": {
                                        "user": {"$ref": "#/components/schemas/userCreate"},
                                        "signup": {"$ref": "#/components/schemas/signup"},
                                    },
                                },
                            }
                        },
                        "parameters": {
                            "userId": {
                                "description": "Id of the user.",
                                "schema": {"type": "string"},
                            }
                        },
                        "correlationIds": {
                            "default": {
                                "description": "Default Correlation ID",
                                "location": "$message.header#/correlationId",
                            }
                        },
                        "messageTraits": {
                            "commonHeaders": {
                                "headers": {
                                    "type": "object",
                                    "properties": {
                                        "my-app-header": {
                                            "type": "integer",
                                            "minimum": 0,
                                            "maximum": 100,
                                        }
                                    },
                                }
                            }
                        },
                    }
                }
            ]
        }



from typing import Optional, Union

from pydantic import Extra, Field

from aiosox.utils import BaseModel

from .reference import Reference
from .schema import ApiSchema

ApiParameterName = str


class ApiParameter(BaseModel):
    """Describes a parameter included in a channel name."""

    description: Optional[str] = None
    """
    A verbose explanation of the parameter. CommonMark syntax can be used for rich
    text representation.
    """

    param_schema: Optional[Union[ApiSchema, Reference]] = Field(default=None, alias="schema")
    """
    Definition of the parameter.
    """

    location: Optional[str] = None
    """
    A runtime expression that specifies the location of the parameter value. Even
    when a definition for the target field exists, it MUST NOT be used to validate
    this parameter but, instead, the schema property MUST be used.
    """

    class Config:
        extra = Extra.forbid
        schema_extra = {
            "examples": [
                {
                    "user/{userId}/signup": {
                        "parameters": {
                            "userId": {
                                "description": "Id of the user.",
                                "schema": {"type": "string"},
                                "location": "$message.payload#/user/id",
                            }
                        },
                        "subscribe": {"message": {"$ref": "#/components/messages/userSignedUp"}},
                    }
                }
            ]
        }



from typing import Dict, List, Mapping, Optional, Type

import orjson
from fastapi.encoders import jsonable_encoder
from pydantic import AnyUrl, Extra, Field, schema_json_of, schema_of
 
from aiosox.utils import BaseModel

from .channel import ApiChannel, ApiOperation
from .components import ApiComponents
from .message import ApiMessage
from .reference import Reference
from .schema import ApiSchema
from .server import ApiServer, ApiServerName
from .tag import ApiTag

 
ChannelName = str


class License(BaseModel):
    """
    License information for the exposed API.
    """

    name: str
    url: Optional[AnyUrl] = None

    class Config:
        extra = Extra.forbid


class Contact(BaseModel):
    """
    Contact information for the exposed API.
    """

    name: Optional[str] = None
    url: Optional[AnyUrl] = None
    email: Optional[str] = None

    class Config:
        extra = Extra.forbid


class ApiInfo(BaseModel):
    """base data for app api"""

    title: str = Field(..., description="app api title")
    version: str = Field(..., description="app api version")
    description: Optional[str] = Field(None, description="description of app api")
    terms_of_service: Optional[AnyUrl] = None
    contact: Optional[Contact] = None
    license: Optional[License] = None


class AsyncAPI(BaseModel):
    """combines all the parts of the asyncapi schema"""

    """base data for asyncapi spec"""
    id: str | None = None
    info: ApiInfo = Field(description="api base data")
    asyncapi: str = Field(default="2.5.0", description="version of the asyncapi spec")
    servers: Mapping[ApiServerName, ApiServer]
    channels: Mapping[ChannelName, ApiChannel] = {}
    components: ApiComponents
    tags: List[ApiTag]

 
   

 


from typing import Any, Generic, List, Optional, Type, TypeVar, Union

from fastapi.encoders import jsonable_encoder
from socketio import AsyncServer

from .enums import SioAuth

from ..utils import BaseModel
from .schemas import SioError

T = TypeVar("T", bound=BaseModel)


class SIOActorMeta(BaseModel):
    event: str
    title: str | None
    summary: str | None
    description: str | None
    model: Type[BaseModel] | Union[Any, Type[BaseModel]] | None = None
    media_type: str
    message_description: str | None
    namespace: str
    auth:SioAuth 

class SIOEmitterMeta(SIOActorMeta):

    include: Optional[Any] = None
    exclude: Optional[Any] = None
    by_alias: bool = True
    exclude_unset: bool = False
    exclude_defaults: bool = False
    exclude_none: bool = False


class SIOJsonEmitter(Generic[T]):
    def __init__(self, model: Type[T]|Union[Type[T], Type[T]]| List[Type[T]], meta: SIOEmitterMeta, sio: AsyncServer):
        self._meta = meta
        self._model = model
        self._sio = sio

    def get_meta(self):
        return self._meta

    async def emit(self, payload: T, emit_to: List[str], encode_kwargs={}, **kwargs):
        meta_args = self._meta.dict(
            include={
                "include",
                "exclude",
                "by_alias",
                "exclude_unset",
                "exclude_defaults",
                "exclude_none",
            }
        )

        await self._sio.emit(
            self._meta.event,
            data=jsonable_encoder(payload, **(meta_args | encode_kwargs)),
            sid=emit_to,
            namespace="/" + self._meta.namespace,
        )


class SIOHandler(SIOActorMeta):
    name: str | None
    response_model: Type[BaseModel] | Union[Any, Type[BaseModel]] | List[
        Type[BaseModel]
    ] | None = None
    


from types import UnionType
from typing import Dict, List, Type, TypeVar, Union, get_origin

import socketio
from fastapi import FastAPI
from pydantic import BaseModel
from pydantic.schema import schema

from aiosox import asyncapi

from ..asyncapi.bindings import (
    WebSocketsOperationBinding,
)
from ..asyncapi.message import ApiMessage
from ..asyncapi.operation import OperationBindings
from ..utils import to_camel, to_no_underscore
from .actors import SIOHandler, SIOJsonEmitter
from .client_manager import KafkaBackend, SocketIoClientManager
from .enums import SioAuth
from .schemas import SioError

T = TypeVar("T", bound=BaseModel)


def resolve_model(
    model: Type[BaseModel] | Union[Type[BaseModel], Type[BaseModel]] | List[Type[BaseModel]]
):
    """method to determine the schema model,type and ref"""
    origin = get_origin(model)
    # standard
    if origin is None:
        return model
    # union
    if origin == Union or origin == UnionType:
        return [resolve_model(m) for m in list(model.__args__)]

    # list
    elif issubclass(origin, list):
        return resolve_model(list(model.__args__)[0])


def resolve_api_message(actor: SIOJsonEmitter | SIOHandler) -> ApiMessage:
    """method to determine api message from a handler/emmit"""

    # handler
    if issubclass(type(actor), SIOHandler):

        name = actor.event
        content_type = actor.media_type
        payload = {"oneOf": []}
        # payload model
        if actor.model is not None:
            resolved_model = resolve_model(actor.model)
            # union
            if issubclass(type(resolved_model), list):
                payload = {
                    "oneOf": [
                        {"$ref": f"{asyncapi.SCHEMA_REF_PREFIX}{m.__name__}"}
                        for m in resolved_model
                    ]
                }
            else:
                model_origin = None
                model_origin = get_origin(actor.model)
                # list
                if model_origin and issubclass(model_origin, list):
                    payload["oneOf"].append(
                        {
                            "type": "array",
                            "items": {
                                "$ref": f"{asyncapi.SCHEMA_REF_PREFIX}{actor.model.__args__[0].__name__}"
                            },
                        }
                    )
                # single model
                else:
                    payload["oneOf"].append(
                        {"$ref": f"{asyncapi.SCHEMA_REF_PREFIX}{actor.model.__name__}"}
                    )
        # response model
        x_response = {"oneOf": []}
        if actor.response_model is not None:
            resolved_response_model = resolve_model(actor.response_model)
            # union
            if issubclass(type(resolved_response_model), list):
                x_response = {
                    "oneOf": [
                        {"$ref": f"{asyncapi.SCHEMA_REF_PREFIX}{m.__name__}"}
                        for m in resolved_response_model
                    ]
                }
            else:
                model_origin = None
                model_origin = get_origin(actor.response_model)
                # list
                if model_origin and issubclass(model_origin, list):
                    x_response["oneOf"].append(
                        {
                            "type": "array",
                            "items": {
                                "$ref": f"{asyncapi.SCHEMA_REF_PREFIX}{actor.response_model.__args__[0].__name__}"
                            },
                        }
                    )
                # single model
                else:
                    x_response["oneOf"].append(
                        {"$ref": f"{asyncapi.SCHEMA_REF_PREFIX}{actor.response_model.__name__}"}
                    )

        description = actor.message_description
        x_response["oneOf"].append({"$ref": f"{asyncapi.SCHEMA_REF_PREFIX}{SioError.__name__}"})
    # emmiter
    elif issubclass(type(actor), SIOJsonEmitter):
        name = actor._meta.event
        payload = {"oneOf": []}
        description = None
        content_type = actor._meta.media_type
        x_response = None
        if actor._model is not None:
            resolved_response_model = resolve_model(actor._model)

            # union
            if issubclass(type(resolved_response_model), list):
                payload = {
                    "oneOf": [
                        {"$ref": f"{asyncapi.SCHEMA_REF_PREFIX}{m.__name__}"}
                        for m in resolved_response_model
                    ]
                }
            else:
                model_origin = None
                model_origin = get_origin(actor._model)
                # list
                if model_origin and issubclass(model_origin, list):
                    payload["oneOf"].append(
                        {
                            "type": "array",
                            "items": {
                                "$ref": f"{asyncapi.SCHEMA_REF_PREFIX}{actor._model.__args__[0].__name__}"
                            },
                        }
                    )
                # single model
                else:
                    payload["oneOf"].append(
                        {"$ref": f"{asyncapi.SCHEMA_REF_PREFIX}{actor._model.__name__}"}
                    )

        description = actor._meta.message_description

    else:
        raise Exception("message for actor not supported")
    return asyncapi.ApiMessage(  # type:ignore
        name=name,
        contentType=content_type,
        description=description,
        payload=payload,
        x_response=x_response,  # type:ignore
    )


def resolve_api_opperation_binding(auth: SioAuth) -> OperationBindings:
    """method to resolve which auth to be used for the method"""
    print(auth)
    if str(auth) == SioAuth.token:
        return OperationBindings(
            socketio=WebSocketsOperationBinding(
                headers={"$ref": "#/components/securitySchemes/apiKey"}
            )
        )

    elif auth  == SioAuth.jwt:
        return OperationBindings(
            socketio=WebSocketsOperationBinding(
                headers={"$ref": "#/components/securitySchemes/bearer"}
            )
        )
    else:
        return OperationBindings(socketio=WebSocketsOperationBinding())


class SocketIoServer:
    """socketio server"""

    def __init__(
        self,
        app: FastAPI,
        kafka_url: str,
    ) -> None:

        group_id = app.title.lower() if app.title else None
        kafka_backend: KafkaBackend = KafkaBackend(kafka_url, group_id=group_id)
        self.client_manager: SocketIoClientManager = SocketIoClientManager(
            kafka_backend=kafka_backend
        )

        self._sio: socketio.AsyncServer = socketio.AsyncServer(
            async_mode="asgi",
            cors_allowed_origins="*",
            cors_credentials=True,
            SameSite=None,
            monitor_clients=True,
            client_manager=self.client_manager,
        )

        self._asgiapp: socketio.ASGIApp = socketio.ASGIApp(
            socketio_server=self._sio,
        )
        self._app: FastAPI = app
        self.info = asyncapi.ApiInfo(
            title=self._app.title,
            version=self._app.openapi_version,
            description=self._app.description,
        )
        self.servers = {
            "dev": asyncapi.ApiServer(
                url="localhost:8000",
                protocol=asyncapi.ApiServerProtocol.SIO,
                security=[{"apiKey": []}, {"bearer": []}],  # type:ignore
            )
        }

        self.tags = []

        self.asyncapi_schema: asyncapi.AsyncAPI | None = None
        self._app.get(
            "/asyncapi.json",
            include_in_schema=False,
            response_model=asyncapi.AsyncAPI,
            response_model_exclude_none=True,
        )(self._asyncapi)

        self._app.mount("/ws", self._asgiapp)

        self._handlers: List[SIOHandler] = []
        self._emitters: List[SIOJsonEmitter] = []

    def _generate_operation_id_for_method(self, event_name: str, namespace: str) -> str | None:
        """method to generate a unique id for each method event of the namspace"""

        return to_camel(
            f"{namespace[1:].capitalize()}{to_no_underscore(event_name.capitalize()).replace(' ','')}"
        )

    def _asyncapi(self) -> "asyncapi.AsyncAPI":
        if not self.asyncapi_schema:

            models_to_resolve = (
                [handler.model for handler in self._handlers if handler.model is not None]
                + [emitter._model for emitter in self._emitters if emitter._model is not None]
                + [
                    handler.response_model
                    for handler in self._handlers
                    if handler.response_model is not None
                ]
                + [SioError]
            )
            resolved_models = []
            for m in models_to_resolve:
                resolved = resolve_model(m)
                if type(resolved) is list and not type(resolved) == Type[BaseModel]:
                    [resolved_models.append(r) for r in resolved]

                else:
                    resolved_models.append(resolved)

            used_models = resolved_models
            components = self.get_components(used_models)
            self.asyncapi_schema = asyncapi.AsyncAPI(
                servers=self.servers,
                tags=self.tags,
                info=self.info,
                components=components,
                id="urn:com:" + "_".join(self._app.title.lower().split(" ")),
                channels=self.get_channels(),
            )

        return self.asyncapi_schema

    def get_components(self, used_models: List[Type[BaseModel]]):
        return asyncapi.ApiComponents(
            schemas=schema(used_models, ref_prefix=asyncapi.SCHEMA_REF_PREFIX)["definitions"],
            security_schemes={
                "apiKey": asyncapi.ApiSecurityScheme(  # type:ignore
                    param_type=asyncapi.ApiSecuritySchemesType.HTTP_API_KEY,  # type:ignore
                    name="token",
                    param_in=asyncapi.ApiSecuritySchemeLocation.header,  # type:ignore
                ),
                "bearer": asyncapi.ApiSecurityScheme(  # type:ignore
                    param_type=asyncapi.ApiSecuritySchemesType.HTTP,  # type:ignore
                    scheme="bearer",
                    bearerFormat="Bearer",
                ),
            },
        )

    def get_channels(
        self,
    ) -> Dict[str, asyncapi.ApiChannel]:

        return {
            f"{handler.namespace}/"
            + handler.event: asyncapi.ApiChannel(  # type:ignore
                publish=asyncapi.ApiOperation(
                    operationId=handler.name,
                    summary=handler.summary,
                    description=handler.description,
                    message=resolve_api_message(handler),
                    bindings=resolve_api_opperation_binding(handler.auth),
                ),
            )
            for handler in self._handlers
        } | {
            f"{emitter._meta.namespace}/"
            + emitter._meta.event: asyncapi.ApiChannel(  # type:ignore
                subscribe=asyncapi.ApiOperation(
                    operationId=self._generate_operation_id_for_method(
                        namespace=emitter._meta.namespace, event_name=emitter._meta.event
                    ),
                    summary=emitter._meta.summary,
                    description=emitter._meta.description,
                    message=resolve_api_message(emitter),
                    bindings=resolve_api_opperation_binding(emitter._meta.auth),
                ),
            )
            for emitter in self._emitters
        }

    async def start(self):
        """on init"""
        await self.client_manager.on_start()

    async def shutdown(self):
        """on shutdown"""
        await self.client_manager.on_shutdown()

    def is_asyncio_based(self) -> bool:
        return True

    @property
    def attach(self):
        return self._sio.attach

    @property
    def emit(self):
        return self._sio.emit

    @property
    def send(self):
        return self._sio.send

    @property
    def call(self):
        return self._sio.call

    @property
    def close_room(self):
        return self._sio.close_room

    @property
    def get_session(self):
        return self._sio.get_session

    @property
    def save_session(self):
        return self._sio.save_session

    @property
    def session(self):
        return self._sio.session

    @property
    def disconnect(self):
        return self._sio.disconnect

    @property
    def handle_request(self):
        return self._sio.handle_request

    @property
    def start_background_task(self):
        return self._sio.start_background_task

    @property
    def sleep(self):
        return self._sio.sleep

    @property
    def enter_room(self):
        return self._sio.enter_room

    @property
    def leave_room(self):
        return self._sio.leave_room

    @property
    def register_namespace(self):

        return self._sio.register_namespace



from enum import StrEnum, auto


class SioAuth(StrEnum):
    """determines if the channel subs/pub requires the client to be authenticated and what type of auth is it"""

    no_auth = auto()
    jwt = auto()
    token = auto()



from enum import StrEnum, auto

from aiosox.utils import CamelStrEnum


class ApiServerProtocol(StrEnum):

    WS = auto()
    WSS = auto()
    HTTP = auto()
    HTTPS = auto()
    SIO = auto()


class ApiSecuritySchemesType(CamelStrEnum):
    """
    https://www.asyncapi.com/docs/specifications/v2.3.0#securitySchemeObjectType
    """

    USER_PASSWORD = auto()
    API_KEY = auto()
    X509 = auto()
    SYMMETRIC_ENCRYPTION = auto()
    ASYMMETRIC_ENCRYPTION = auto()
    HTTP_API_KEY = auto()
    HTTP = auto()
    OAUTH2 = auto()
    OPENID_CONNECT = auto()
    PLAIN = auto()
    SCRAM_SHA256 = auto()
    SCRAM_SHA512 = auto()
    GSSAPI = auto()


class ApiSecuritySchemeLocation(CamelStrEnum):
    """
    https://www.asyncapi.com/docs/specifications/v2.3.0#securitySchemeObject
    """

    user = auto()
    password = auto()
    apiKey = auto()
    query = auto()
    header = auto()
    cookie = auto()
    httpApiKey = auto()


class JSONReference(StrEnum):
    LOCAL = auto()
    REMOTE = auto()
    URL = auto()


class HttpOperationBindingType(StrEnum):
    request = auto()
    response = auto()


class WebSocketsMethod(StrEnum):
    get = auto()
    post = auto()




import orjson
from aiokafka import AIOKafkaConsumer, AIOKafkaProducer
from kafka.coordinator.assignors.roundrobin import RoundRobinPartitionAssignor


def kafka_serialize(value):
    """function to handle encoding of a value"""
    if value is None:
        return None
    result = None
    try:
        result = orjson.dumps(value)
    except orjson.JSONEncodeError as e:
        return None
    return result


def kafka_deserialize(value):
    """function to handle decoding of a value"""
    if not isinstance(value, bytes):
        return orjson.loads(orjson.dumps(value))
    decoded_byte_value = value.decode("utf-8")
    return orjson.loads(orjson.dumps(decoded_byte_value))


def get_kafka_consumer(
    *topics,
    bootstrap_servers="localhost",
    group_id: str|None,
    key_deserializer=kafka_deserialize,
    value_deserializer=kafka_deserialize,
    fetch_max_wait_ms=500,
    fetch_max_bytes=52428800,
    fetch_min_bytes=1,
    max_partition_fetch_bytes=1 * 1024 * 1024,
    request_timeout_ms=40 * 1000,
    retry_backoff_ms=100,
    auto_offset_reset="latest",
    enable_auto_commit=True,
    auto_commit_interval_ms=5000,
    check_crcs=True,
    metadata_max_age_ms=5 * 60 * 1000,
    partition_assignment_strategy=(RoundRobinPartitionAssignor,),
    max_poll_interval_ms=300000,
    rebalance_timeout_ms=None,
    session_timeout_ms=10000,
    heartbeat_interval_ms=3000,
    consumer_timeout_ms=200,
    max_poll_records=None,
    ssl_context=None,
    security_protocol="PLAINTEXT",
    api_version="auto",
    exclude_internal_topics=True,
    connections_max_idle_ms=540000,
    isolation_level="read_uncommitted",
    sasl_mechanism="PLAIN",
    sasl_plain_password=None,
    sasl_plain_username=None,
    sasl_kerberos_service_name="kafka",
    sasl_kerberos_domain_name=None,
    sasl_oauth_token_provider=None,
    ssl_cert=None,
) -> AIOKafkaConsumer:
    """function to get an instance of aiokafka consumer"""

    return AIOKafkaConsumer(
        *topics,
        value_deserializer=value_deserializer,
        bootstrap_servers=bootstrap_servers,
        auto_offset_reset=auto_offset_reset,
        isolation_level=isolation_level,
        group_id=group_id,
        key_deserializer=key_deserializer,
        fetch_max_wait_ms=fetch_max_wait_ms,
        fetch_max_bytes=fetch_max_bytes,
        fetch_min_bytes=fetch_min_bytes,
        max_partition_fetch_bytes=max_partition_fetch_bytes,
        request_timeout_ms=request_timeout_ms,
        retry_backoff_ms=retry_backoff_ms,
        enable_auto_commit=enable_auto_commit,
        auto_commit_interval_ms=auto_commit_interval_ms,
        check_crcs=check_crcs,
        metadata_max_age_ms=metadata_max_age_ms,
        partition_assignment_strategy=partition_assignment_strategy,
        max_poll_interval_ms=max_poll_interval_ms,
        rebalance_timeout_ms=rebalance_timeout_ms,
        session_timeout_ms=session_timeout_ms,
        heartbeat_interval_ms=heartbeat_interval_ms,
        consumer_timeout_ms=consumer_timeout_ms,
        max_poll_records=max_poll_records,
        ssl_context=ssl_context,
        security_protocol=security_protocol,
        api_version=api_version,
        exclude_internal_topics=exclude_internal_topics,
        connections_max_idle_ms=connections_max_idle_ms,
        sasl_mechanism=sasl_mechanism,
        sasl_plain_password=sasl_plain_password,
        sasl_plain_username=sasl_plain_username,
        sasl_kerberos_service_name=sasl_kerberos_service_name,
        sasl_kerberos_domain_name=sasl_kerberos_domain_name,
        sasl_oauth_token_provider=sasl_oauth_token_provider,
    )


def get_kafka_producer(
    bootstrap_servers: str,
) -> AIOKafkaProducer:
    security_protocol = "PLAINTEXT"

    return AIOKafkaProducer(
        bootstrap_servers=bootstrap_servers,
        value_serializer=kafka_serialize,
        security_protocol=security_protocol,
    )



from aiosox.utils import snake2camel, title_case, to_camel

from .main import *
from .enums import *
from .components import ApiComponents
from .security import ApiSecurityScheme

from .channel import ApiChannel,ApiOperation
from .reference import Reference
SCHEMA_REF_PREFIX = "#/components/schemas/"



from aiosox import utils


class Test_Utils_To_no_underscore:
    def test_to_no_underscore_1(self):
        utils.to_no_underscore("2017-09-29T23:01:00.000Z")

    def test_to_no_underscore_2(self):
        utils.to_no_underscore("2017-09-29T19:01:00.000")

    def test_to_no_underscore_3(self):
        utils.to_no_underscore("01:04:03")

    def test_to_no_underscore_4(self):
        utils.to_no_underscore("Mon Aug 03 12:45:00")

    def test_to_no_underscore_5(self):
        utils.to_no_underscore("")

    def test_to_no_underscore_6(self):
        utils.to_no_underscore(None)


class Test_Utils_Title_case:
    def test_title_case_1(self):
        result = utils.title_case("Mon Aug 03 12:45:00")

    def test_title_case_2(self):
        result = utils.title_case("2017-09-29T23:01:00.000Z")

    def test_title_case_3(self):
        result = utils.title_case("01:04:03")

    def test_title_case_4(self):
        result = utils.title_case("")



from typing import Optional

from pydantic import Extra, Field

from aiosox.utils import BaseModel

from .enums import HttpOperationBindingType, WebSocketsMethod
from .schema import ApiSchema


class HttpChannelBinding(BaseModel):
    """
    This document defines how to describe HTTP-specific information on AsyncAPI.
    This object MUST NOT contain any properties. Its name is reserved for future use.
    """

    class Config:
        extra = Extra.forbid


class HttpMessageBinding(BaseModel):
    """
    This object contains information about the message representation in HTTP.
    """

    headers: Optional[ApiSchema] = None
    """
    A Schema object containing the definitions for HTTP-specific headers. This
    schema MUST be of type object and have a properties key.
    """

    bindingVersion: Optional[str] = None
    """
    The version of this binding. If omitted, "latest" MUST be assumed.
    """

    class Config:
        extra = Extra.forbid


class HttpServerBinding(BaseModel):
    """
    This document defines how to describe HTTP-specific information on AsyncAPI.
    This object MUST NOT contain any properties. Its name is reserved for future use.
    """

    class Config:
        extra = Extra.forbid


class HttpOperationBinding(BaseModel):
    """
    This document defines how to describe HTTP-specific information on AsyncAPI.
    """

    param_type: HttpOperationBindingType = Field(alias="type")
    """
    **REQUIRED**. Type of operation. Its value MUST be either request or response.
    """

    method: Optional[str] = None
    """
    When type is request, this is the HTTP method, otherwise it MUST be ignored.
    Its value MUST be one of GET, POST, PUT, PATCH, DELETE, HEAD, OPTIONS, CONNECT, and TRACE.
    """

    query: Optional[ApiSchema] = None
    """
    A Schema object containing the definitions for each query parameter.
    This schema MUST be of type object and have a properties key.
    """

    bindingVersion: Optional[str] = None
    """
    The version of this binding. If omitted, "latest" MUST be assumed.
    """

    class Config:
        extra = Extra.forbid
        schema_extra = {
            "examples": [
                {
                    "type": "request",
                    "method": "GET",
                    "query": {
                        "type": "object",
                        "required": ["companyId"],
                        "properties": {
                            "companyId": {
                                "type": "number",
                                "minimum": 1,
                                "description": "The Id of the company.",
                            }
                        },
                        "additionalProperties": False,
                    },
                    "bindingVersion": "0.1.0",
                }
            ]
        }


class WebSocketsChannelBinding(BaseModel):
    """
    When using WebSockets, the channel represents the connection. Unlike other
    protocols that support multiple virtual channels (topics, routing keys, etc.)
    per connection, WebSockets doesn't support virtual channels or, put it another
    way, there's only one channel and its characteristics are strongly related to
    the protocol used for the handshake, i.e., HTTP.
    """

    method: Optional[WebSocketsMethod] = None
    """
    The HTTP method to use when establishing the connection. Its value MUST be
    either GET or POST.
    """

    query: Optional[ApiSchema] = None
    """
    A Schema object containing the definitions for each query parameter. This
    schema MUST be of type object and have a properties key.
    """

    headers: Optional[ApiSchema] = None
    """
    A Schema object containing the definitions of the HTTP headers to use when
    establishing the connection. This schema MUST be of type object and have a
    properties key.
    """

    bindingVersion: Optional[str] = None
    """
    The version of this binding. If omitted, "latest" MUST be assumed.
    """


class WebSocketsMessageBinding(BaseModel):
    """
    This document defines how to describe WebSockets-specific information on AsyncAPI.
    This object MUST NOT contain any properties. Its name is reserved for future use.
    """

    class Config:
        extra = Extra.forbid


class WebSocketsOperationBinding(BaseModel):
    """
    This document defines how to describe WebSockets-specific information on AsyncAPI.
    This object MUST NOT contain any properties. Its name is reserved for future use.
    """
    headers: Optional[ApiSchema] = None
    """
    A Schema object containing the definitions of the HTTP headers to use when
    establishing the connection. This schema MUST be of type object and have a
    properties key.
    """
    class Config:
        extra = Extra.forbid


class WebSocketsServerBinding(BaseModel):
    """
    This document defines how to describe WebSockets-specific information on AsyncAPI.
    This object MUST NOT contain any properties. Its name is reserved for future use.
    """

    class Config:
        extra = Extra.forbid


class KafkaChannelBinding(BaseModel):
    """
    This document defines how to describe Kafka-specific information on AsyncAPI.
    This object MUST NOT contain any properties. Its name is reserved for future use.
    """

    class Config:
        extra = Extra.forbid


class KafkaMessageBinding(BaseModel):
    """
    This object contains information about the message representation in Kafka.
    """

    key: Optional[ApiSchema] = None
    """
    The message key. NOTE: You can also use the reference object way.
    """

    bindingVersion: Optional[str] = None
    """
    The version of this binding. If omitted, "latest" MUST be assumed.
    """

    class Config:
        extra = Extra.forbid


class KafkaServerBinding(BaseModel):
    """
    This document defines how to describe Kafka-specific information on AsyncAPI.
    This object MUST NOT contain any properties. Its name is reserved for future use.
    """

    class Config:
        extra = Extra.forbid


class KafkaOperationBinding(BaseModel):
    """
    This document defines how to describe Kafka-specific information on AsyncAPI.
    """

    groupId: Optional[ApiSchema] = None
    """
    Id of the consumer group.
    """

    clientId: Optional[ApiSchema] = None
    """
    Id of the consumer inside a consumer group.
    """

    bindingVersion: Optional[str] = None
    """
    The version of this binding. If omitted, "latest" MUST be assumed.
    """

    class Config:
        extra = Extra.forbid


class ChannelBindings(BaseModel):
    """
    Map describing protocol-specific definitions for a channel.
    """

    http: Optional[HttpChannelBinding] = None
    """
    Protocol-specific information for an HTTP channel.
    """

    ws: Optional[WebSocketsChannelBinding] = None
    """
    Protocol-specific information for a WebSockets channel.
    """
    socketio: Optional[WebSocketsChannelBinding] = None
    """
    Protocol-specific information for a Socket Io channel.
    """

    kafka: Optional[KafkaChannelBinding] = None

    class Config:
        extra = Extra.forbid


class ApiServerBindings(BaseModel):
    """
    Map describing protocol-specific definitions for a server.
    """

    http: Optional[HttpServerBinding] = None
    """
    Protocol-specific information for an HTTP server.
    """

    ws: Optional[WebSocketsServerBinding] = None
    """
    Protocol-specific information for a WebSockets server.
    """

    kafka: Optional[KafkaServerBinding] = None
    """
    Protocol-specific information for a Kafka server.
    """

    class Config:
        extra = Extra.forbid



import pytest

from aiosox.sio import namespace


class Test_Sionamespace_On:
    @pytest.fixture()
    def sionamespace(self):
        return namespace.SioNamespace()

    def test_on_1(self, sionamespace):
        result = sionamespace.on(
            "data:image/svg+xml;charset=UTF-8,%3Csvg%20xmlns%3D%22http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg%22%20version%3D%221.1%22%20baseProfile%3D%22full%22%20width%3D%22undefined%22%20height%3D%22undefined%22%3E%3Crect%20width%3D%22100%25%22%20height%3D%22100%25%22%20fill%3D%22grey%22%2F%3E%3Ctext%20x%3D%22NaN%22%20y%3D%22NaN%22%20font-size%3D%2220%22%20alignment-baseline%3D%22middle%22%20text-anchor%3D%22middle%22%20fill%3D%22white%22%3Eundefinedxundefined%3C%2Ftext%3E%3C%2Fsvg%3E"
        )

    def test_on_2(self, sionamespace):
        result = sionamespace.on("")



from typing import TYPE_CHECKING, Any, Callable, List, Optional, Type, TypeVar, Union

import socketio
from pydantic import BaseModel

from ..utils import to_camel
from .actors import SIOEmitterMeta, SIOHandler, SIOJsonEmitter
from .enums import SioAuth

if TYPE_CHECKING:
    from .server import SocketIoServer

T = TypeVar("T", bound=BaseModel)

Y = TypeVar("Y", bound=BaseModel)


class RequestValidationError(Exception):
    pass


class ResponseValidationError(Exception):
    pass


class SioNamespace(socketio.AsyncNamespace):
    def __init__(
        self,
        namespace: str,
        socket_io_server: "SocketIoServer",
    ) -> None:
        self._namespace_name = namespace
        self._socket_io_server: "SocketIoServer" = socket_io_server  # type:ignore
        super(socketio.Namespace, self).__init__(namespace=namespace)

    def on(
        self,
        event: str,
        title: str | None = None,
        summary: str | None = None,
        description: str | None = None,
        message_description: str | None = None,
        payload_model: Type[BaseModel] | Union[Type[BaseModel], Any] | None = None,
        response_model: Type[BaseModel]
        | Union[Type[BaseModel], Any]
        | List[Type[BaseModel]]
        | None = None,
        media_type: str = "application/json",
        auth: SioAuth = SioAuth.no_auth,
    ):
        def decorator(fn: Callable):
            self._socket_io_server._handlers.append(
                SIOHandler(
                    name=self._socket_io_server._generate_operation_id_for_method(
                        namespace=self._namespace_name, event_name=fn.__name__
                    ),
                    event=event,
                    title=title,
                    summary=summary,
                    description=description,
                    model=payload_model,
                    response_model=response_model,
                    media_type=media_type,
                    message_description=message_description,
                    namespace=self._namespace_name[1:],
                    auth=auth,
                )
            )

            return self._socket_io_server._sio.on(
                event=event, handler=fn, namespace=self._namespace_name
            )

        return decorator

    def create_emitter(
        self,
        event: str,
        model: Type[List[T]] | Type[Union[Type[T], Type]] | Type[T],
        title: str | None = None,
        summary: str | None = None,
        description: str | None = None,
        message_description: str | None = None,
        media_type: str = "application/json",
        include: Optional[Any] = None,
        exclude: Optional[Any] = None,
        by_alias: bool = True,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        auth: SioAuth = SioAuth.no_auth,
    ):

        emitter = SIOJsonEmitter(
            model=model,
            meta=SIOEmitterMeta(
                event=to_camel(event),
                title=title,
                summary=summary,
                description=description,
                model=model,
                media_type=media_type,
                message_description=message_description,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                namespace=self._namespace_name[1:],
                auth=auth,
            ),
            sio=self._socket_io_server._sio,
        )
        self._socket_io_server._emitters.append(emitter)
        return emitter



import datetime
import re
import time
from enum import StrEnum
from typing import Any, Callable

import orjson
from pydantic import BaseModel as PydanticBase




def to_camel(string: str) -> str:
    split_string = string.split("_")
    if len(split_string) > 1:
        word_list = [split_string[0].lower()]
        for word in split_string[1:]:
            word_list.append(word.capitalize())
        string = "".join(word_list)
    return string


def snake_case(any_string: str):
    """method to camel case to snake case"""
    any_string = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", any_string)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", any_string).lower()


def title_case(any_string: str) -> str:
    split_string = any_string.split("_")
    if len(split_string) > 1:
        word_list = []
        for word in split_string:
            word_list.append("{word} ".format(word=word.capitalize()))
        any_string = "".join(word_list)
    else:
        any_string = any_string.capitalize()
    return any_string.strip()


def snake2camel(snake: str, start_lower: bool = False) -> str:
    camel = snake.title()
    camel = re.sub("([0-9A-Za-z])_(?=[0-9A-Z])", lambda m: m.group(1), camel)
    if start_lower:
        camel = re.sub("(^_*[A-Z])", lambda m: m.group(1).lower(), camel)
    return camel


def to_no_underscore(string: str) -> str:
    if string:
        split_string = string.split("_")
        if len(split_string) > 1:
            word_list = []
            for word in split_string:
                word_list.append("{word} ".format(word=word.capitalize()))
            string = "".join(word_list)
        else:
            string = string.capitalize()
    return string


class CamelStrEnum(StrEnum):
    """
    CamelStrEnum subclasses that create variants using `auto()` will have values equal to their camelCase names
    """

    # noinspection PyMethodParameters
    def _generate_next_value_(name, start, count, last_values) -> str:  # type: ignore
        """
        Uses the camelCase name as the automatic value, rather than an integer
        See https://docs.python.org/3/library/enum.html#using-automatic-values for reference
        """
        return snake2camel(name, start_lower=True)


class BaseModel(PydanticBase):
    class Config:
        """base pydantic config"""

        alias_generator = to_camel
        arbitrary_types_allowed = True
        use_enum_values = True
        json_loads = orjson.loads
        json_dumps = orjson.dumps
        allow_population_by_field_name = True
        smart_union = True
        json_encoders = {
            datetime.datetime: lambda v: round(v.timestamp()) * 1000,
            datetime.date: lambda v: round(time.mktime(v.timetuple())) * 1000,
        }

 

 
  


from typing import Any, Dict, List, Mapping, Optional, Sequence, Set, Type, Union

from fastapi.encoders import jsonable_encoder
from pydantic import AnyUrl, Extra, Field, constr, root_validator, validator

from aiosox.utils import BaseModel

from .bindings import (
    HttpChannelBinding,
    HttpMessageBinding,
    KafkaChannelBinding,
    KafkaMessageBinding,
    WebSocketsChannelBinding,
    WebSocketsMessageBinding,
)
from .enums import (
    ApiSecuritySchemeLocation,
    ApiSecuritySchemesType,
    ApiServerProtocol,
    JSONReference,
)
from .reference import Reference
from .schema import ApiSchema, JsonSchemaObject
from .tag import ApiTag


class CorrelationId(BaseModel):
    """
    An object that specifies an identifier at design time that can used for message tracing
    and correlation. For specifying and computing the location of a Correlation ID, a runtime
    expression is used.
    This object can be extended with Specification Extensions.
    """

    description: Optional[str] = None
    """
    An optional description of the identifier. CommonMark syntax can be used for rich text
    representation.
    """

    location: str
    """
    **REQUIRED**. A runtime expression that specifies the location of the correlation ID.
    """

    class Config:
        extra = Extra.forbid
        schema_extra = {
            "examples": [
                {
                    "description": "Default Correlation ID",
                    "location": "$message.header#/correlationId",
                }
            ]
        }


class MessageBindings(BaseModel):
    """
    Map describing protocol-specific definitions for a message.
    """

    http: Optional[HttpMessageBinding] = None
    """
    Protocol-specific information for an HTTP message, i.e., a request or a response.
    """

    ws: Optional[WebSocketsMessageBinding] = None
    """
    Protocol-specific information for a WebSockets message.
    """

    kafka: Optional[KafkaMessageBinding] = None


class MessageExample(BaseModel):
    """
    Message Example Object represents an example of a Message Object and MUST
    contain either headers and/or payload fields.
    """

    headers: Optional[Dict[str, Union[ApiSchema, Reference]]] = None
    """
    The value of this field MUST validate against the Message Object's headers field.
    """

    payload: Optional[ApiSchema] = None
    """
    The value of this field MUST validate against the Message Object's payload field.
    """

    name: Optional[str] = None
    """
    A machine-friendly name.
    """

    summary: Optional[str] = None
    """
    A short summary of what the example is about.
    """

    class Config:
        extra = Extra.forbid
        schema_extra = {
            "examples": [
                {
                    "name": "SimpleSignup",
                    "summary": "A simple UserSignup example message",
                    "headers": {
                        "correlationId": "my-correlation-id",
                        "applicationInstanceId": "myInstanceId",
                    },
                    "payload": {
                        "user": {"someUserKey": "someUserValue"},
                        "signup": {"someSignupKey": "someSignupValue"},
                    },
                }
            ]
        }


class MessageTrait(BaseModel):
    """
    Describes a trait that MAY be applied to a Message Object. This object MAY contain
    any property from the Message Object, except payload and traits.
    If you're looking to apply traits to an operation, see the Operation Trait Object.
    """

    headers: Optional[Union[ApiSchema, Reference]] = None
    """
    Schema definition of the application headers. Schema MUST be of type "object". It
    MUST NOT define the protocol headers.
    """

    correlationId: Optional[Union[CorrelationId, Reference]] = None
    """
    Definition of the correlation ID used for message tracing or matching.
    """

    schemaFormat: Optional[str] = None
    """
    A string containing the name of the schema format/language used to define the message
    payload. If omitted, implementations should parse the payload as a Schema object.
    """

    contentType: Optional[str] = None
    """
    The content type to use when encoding/decoding a message's payload. The value MUST be
    a specific media type (e.g. application/json). When omitted, the value MUST be the
    one specified on the defaultContentType field.
    """

    name: Optional[str] = None
    """
    A machine-friendly name for the message.
    """

    title: Optional[str] = None
    """
    A human-friendly title for the message.
    """

    summary: Optional[str] = None
    """
    A short summary of what the message is about.
    """

    description: Optional[str] = None
    """
    A verbose explanation of the message. CommonMark syntax can be used for rich text
    representation.
    """

    tags: Optional[List[ApiTag]] = None
    """
    A list of tags for API documentation control. Tags can be used for logical grouping
    of operations.
    """

    bindings: Optional[Union[MessageBindings, Reference]] = None
    """
    A map where the keys describe the name of the protocol and the values describe
    protocol-specific definitions for the message.
    """

    examples: Optional[List[MessageExample]] = None
    """
    List of examples.
    """

    class Config:
        extra = Extra.forbid
        schema_extra = {
            "examples": [
                {
                    "schemaFormat": "application/vnd.apache.avro+json;version=1.9.0",
                    "contentType": "application/json",
                }
            ]
        }


 


class ApiMessage(BaseModel):
    """
    Describes a message received on a given channel and operation.
    """

    headers: Optional[Union[ApiSchema, Reference]] = None

    payload: Optional[ApiSchema] = None

    correlationId: Optional[Union[CorrelationId, Reference]] = None

    schemaFormat: Optional[str] = None

    contentType: Optional[str] = None

    name: Optional[str] = None

    title: Optional[str] = None

    summary: Optional[str] = None

    description: Optional[str] = None

    tags: Optional[List[ApiTag]] = None

    bindings: Optional[Union[MessageBindings, Reference]] = None

    examples: Optional[List[MessageExample]] = None

    traits: Optional[List[Union[MessageTrait, Reference]]] = None

    x_response: Optional[ApiSchema] = Field(alias="x-response")

    class Config:
        extra = Extra.forbid



from typing import Optional

from pydantic import BaseModel, Extra


class ApiTag(BaseModel):
    """Allows adding meta data to a single tag."""

    name: str 
    """
    **REQUIRED**. The name of the tag.
    """

    description: Optional[str] = None
    """
    A short description for the tag. CommonMark syntax can be used for
    rich text representation.
    """

    class Config:
        extra = Extra.forbid



from typing import Dict, List, Optional, Sequence, Union

from pydantic import Field

from aiosox.utils import BaseModel

from .bindings import ChannelBindings
from .operation import ApiOperation
from .parameter import ApiParameter, ApiParameterName
from .reference import Reference
from .security import ApiSecurityRequirement


class ApiChannel(BaseModel):
    """represents a single topic/channel/namespace(socketio)"""

    ref: Optional[str] = Field(alias="$ref")
    description: Optional[str] = None
    servers: Optional[List[str]] = None

    subscribe: Optional[ApiOperation] = None
    publish: Optional[ApiOperation] = None
    parameters: Optional[Dict[ApiParameterName, Union[ApiParameter, Reference]]] = None

    bindings: Optional[Union[ChannelBindings, Reference]] = None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiosox',
 'aiosox.asyncapi',
 'aiosox.kafka',
 'aiosox.ponicode',
 'aiosox.sio',
 'aiosox.sio.ponicode']

package_data = \
{'': ['*']}

install_requires = \
['aiokafka>=0.8.0,<0.9.0',
 'anyio[trio]>=3.6.2,<4.0.0',
 'fastapi>=0.88.0,<0.89.0',
 'mkdocs-material>=9.0.2,<10.0.0',
 'mkdocs>=1.4.2,<2.0.0',
 'orjson>=3.8.3,<4.0.0',
 'pydantic>=1.10.4,<2.0.0',
 'python-socketio>=5.7.2,<6.0.0',
 'uvicorn[standard]>=0.20.0,<0.21.0']

setup_kwargs = {
    'name': 'aiosox',
    'version': '0.1.2',
    'description': '⛓️ Combination of asyncapi(documentation) & socketio pub/sub using aiokafka as the client manager  multinode backend services',
    'long_description': '![Sample image](https://gitlab.com/uploads/-/system/project/avatar/42327849/a5e01db694b47cd07018813ce821a4e1.png?width=64)\n\n\naiosox: <a url="https://gitlab.com/arieutils/aiosox">repo link </a>\n=======================================\nQuick example\n-----------\n\ncan be installed using pip/poetry:\n\n    poetry shell\n    poetry run uvicorn example:app --port=8000\n    \n```python\nfrom typing import List\n\nfrom fastapi import FastAPI\nfrom pydantic import BaseModel\n\nfrom aiosox import SioAuth, SioNamespace, SocketIoServer\n\n\ndef get_app():\n    applitcation = FastAPI(title="tester")\n    return applitcation\n\n\napp = get_app()\n\nsio_server: SocketIoServer = SocketIoServer(app=app, kafka_url="localhost:29092")\nuser_namespapce: SioNamespace = SioNamespace("/user", socket_io_server=sio_server)\nsio_server._sio.register_namespace(user_namespapce)\n\n\n@app.on_event("startup")\nasync def on_start():\n    await sio_server.start()\n\n\n@app.on_event("shutdown")\nasync def on_shutdown():\n    await sio_server.shutdown()\n\n\nclass UserY(BaseModel):\n    name: str\n\n\nclass UserT(BaseModel):\n    name: str\n    what: List[UserY]\n\n\nclass OfferT(BaseModel):\n    title: str\n\n\non_failed_emmiter = user_namespapce.create_emitter("failed", model=OfferT | UserT)\n\n\n@user_namespapce.on(\n    "submit",\n    description="when user submits a form",\n    payload_model=UserT | UserY,\n    response_model=List[UserT],\n    auth=SioAuth.jwt,\n)\nasync def on_submit(sid, data):\n    print(\n        sid,\n        data,\n    )\n\n```',
    'author': 'Arie',
    'author_email': 'ariesorkin3@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gitlab.com/arieutils/aiosox',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '==3.11.1',
}


setup(**setup_kwargs)



from typing import Mapping, Optional, Sequence

from pydantic import Extra, Field

from aiosox.utils import BaseModel

from .enums import ApiSecuritySchemeLocation, ApiSecuritySchemesType

ApiSecurityRequirement = Mapping[str, Sequence[str]]


class ApiSecurityScheme(BaseModel):
    """model defining a security protocol"""

    param_type: ApiSecuritySchemesType = Field(None, alias="type")
    description: Optional[str] = None
    name: Optional[str] = None
    param_in: Optional[ApiSecuritySchemeLocation] = Field(None, alias="in")
    scheme: Optional[str] = None
    bearerFormat: Optional[str] = None

    class Config:
        extra = Extra.forbid



from typing import Dict, List, Optional

from aiosox.utils import BaseModel

from .enums import ApiServerProtocol
from .security import ApiSecurityRequirement

ApiServerName = str


class ApiServer(BaseModel):
    """model defining a server"""

    url: str
    protocol: ApiServerProtocol
    description: Optional[str] = None
    security: Optional[List[Dict[str, ApiSecurityRequirement]]] = None



import asyncio
import typing
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Dict,
    Self,
)

from aiokafka.errors import KafkaConnectionError
from async_generator import asynccontextmanager
from socketio.asyncio_pubsub_manager import AsyncPubSubManager

from ..kafka.utils import get_kafka_consumer, get_kafka_producer


class Unsubscribed(Exception):
    pass


class KafkaEvent:
    def __init__(self, channel: str, message: str) -> None:
        self.channel = channel
        self.message = message

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, KafkaEvent)
            and self.channel == other.channel
            and self.message == other.message
        )

    def __repr__(self) -> str:
        return f"Event(channel={self.channel!r}, message={self.message!r})"


class Subscriber:
    def __init__(self, queue: asyncio.Queue) -> None:
        self._queue = queue

    async def __aiter__(self) -> AsyncGenerator:
        try:
            while True:
                yield await self.get()
        except Unsubscribed:
            pass

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        ...

    async def get(self) -> KafkaEvent:
        item = await self._queue.get()
        if item is None:
            raise Unsubscribed()
        return item


class KafkaBackend:
    def __init__(self, url: str, group_id: str | None = None):
        self._servers = url
        self._consumer_channels: typing.Set = set()
        self.group_id = group_id

    async def connect(self) -> None:
        self._producer = get_kafka_producer(bootstrap_servers=self._servers)
        self._consumer = get_kafka_consumer(bootstrap_servers=self._servers, group_id=self.group_id)

        await self._producer.start()
        await self._consumer.start()

    async def disconnect(self) -> None:
        await self._producer.stop()
        await self._consumer.stop()

    async def subscribe(self, channel: str) -> None:
        self._consumer_channels.add(channel)
        self._consumer.subscribe(topics=self._consumer_channels)

    async def unsubscribe(self, channel: str) -> None:
        self._consumer.unsubscribe()

    async def publish(self, channel: str, message: typing.Any) -> None:
        await self._producer.send_and_wait(channel, message)

    async def next_published(self) -> KafkaEvent:
        message = await self._consumer.getone()
        return KafkaEvent(channel=message.topic, message=message.value)


class SocketIoClientManager(AsyncPubSubManager):
    def __init__(
        self,
        kafka_backend: KafkaBackend,
        channel="socketio",
        write_only=False,
        logger=None,
    ):
        super().__init__(channel, write_only, logger)
        self.kafka_backend: KafkaBackend = kafka_backend
        self._subscribers: Dict[str, Any] = {}
        self._backend = kafka_backend

    async def __aenter__(self) -> Self:
        await self.on_start()
        return self

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        await self.on_shutdown()

    async def _listener(self) -> None:

        while True:
            event = await self._backend.next_published()

            for queue in list(self._subscribers.get(event.channel, [])):
                await queue.put(event)

    async def on_start(self) -> None:
        try:
            await self._backend.connect()
        except KafkaConnectionError as e:
            await self.kafka_backend.disconnect()
            raise RuntimeError("unable to connect to kafka")
        self._listener_task = asyncio.create_task(self._listener())

    async def on_shutdown(self) -> None:
        if self._listener_task.done():
            self._listener_task.result()
        else:
            self._listener_task.cancel()
        await self._backend.disconnect()

    @asynccontextmanager
    async def subscribe(self, channel: str) -> AsyncIterator["Subscriber"]:
        queue: asyncio.Queue = asyncio.Queue()

        try:
            if not self._subscribers.get(channel):
                await self._backend.subscribe(channel)
                self._subscribers[channel] = set([queue])
            else:
                self._subscribers[channel].add(queue)

            yield Subscriber(queue)

            self._subscribers[channel].remove(queue)
            if not self._subscribers.get(channel):
                del self._subscribers[channel]
                await self._backend.unsubscribe(channel)
        finally:
            await queue.put(None)

    async def _publish(self, message: Any):
        await self._backend.publish(self.channel, message)

    async def _listen(self):
        async with self.subscribe(channel=self.channel) as subscriber:  # type:ignore
            async for event in subscriber:
                yield event.message
